@@ -2,7 +2,7 @@ package HTML::Selector::XPath;
 
 use strict;
 use 5.008_001;
-our $VERSION = '0.24';
+our $VERSION = '0.27';
 
 require Exporter;
 our @EXPORT_OK = qw(selector_to_xpath);
@@ -29,7 +29,7 @@ my $reg = {
         (?: ($ident) | "([^"]*)" | '([^']*)') \s* \] /x,
     badattr => qr/^\[/,
     attrN   => qr/^:not\(/i, # we chop off the closing parenthesis below in the code
-    pseudo  => qr/^:([()a-z0-9_+-]+)/i,
+    pseudo  => qr/^:([-\w]+\(?)/i,
     # adjacency/direct descendance
     combinator => qr/^(\s*[>+~\s](?!,))/i,
     # rule separator
@@ -69,15 +69,15 @@ sub convert_attribute_match {
 }
 
 sub _generate_child {
-    my ($direction,$a,$b) = @_;
+    my ($direction,$name,$a,$b) = @_;
     if ($a == 0) { # 0n+b
         $b--;
-        "[count($direction-sibling::*) = $b and parent::*]"
+        "[count($direction-sibling::$name) = $b and parent::*]"
     } elsif ($a > 0) { # an + b
-        return "[not((count($direction-sibling::*)+1)<$b) and ((count($direction-sibling::*) + 1) - $b) mod $a = 0 and parent::*]"
+        return "[not((count($direction-sibling::$name)+1)<$b) and ((count($direction-sibling::$name) + 1) - $b) mod $a = 0 and parent::*]"
     } else { # -an + $b
         $a = -$a;
-        return "[not((count($direction-sibling::*)+1)>$b) and (($b - (count($direction-sibling::*) + 1)) mod $a) = 0 and parent::*]"
+        return "[not((count($direction-sibling::$name)+1)>$b) and (($b - (count($direction-sibling::$name) + 1)) mod $a) = 0 and parent::*]"
     }
 }
 
@@ -86,7 +86,7 @@ sub nth_child {
     if (@_ == 1) {
         ($a,$b) = (0,$a);
     }
-    _generate_child('preceding', $a, $b);
+    _generate_child('preceding', '*', $a, $b);
 }
 
 sub nth_last_child {
@@ -94,11 +94,36 @@ sub nth_last_child {
     if (@_ == 1) {
         ($a,$b) = (0,$a);
     }
-    _generate_child('following', $a, $b);
+    _generate_child('following', '*', $a, $b);
 }
 
 # A hacky recursive descent
 # Only descends for :not(...)
+sub consume_An_plus_B {
+    my( $rrule ) = @_;
+
+    my( $A, $B );
+
+    if( $$rrule =~ s/^odd\s*\)// ) {
+        ($A,$B) = (2, 1)
+    } elsif( $$rrule =~ s/^even\s*\)// ) {
+        ($A,$B) = (2, 0)
+    } elsif( $$rrule =~ s/^\s*(-?\d+)\s*\)// ) {
+        ($A,$B) = (0, $1)
+    } elsif( $$rrule =~ s/^\s*(-?\d*)\s*n\s*(?:\+\s*(\d+))?\s*\)// ) {
+        ($A,$B) = (($1 // '1'), $2 // 0);
+        if( $A eq '-') {
+            $A = '-1';
+        } elsif( $A eq '' ) {
+            $A = '1';
+        }
+    } else {
+        croak "Can't parse formula from '$$rrule'";
+    }
+
+    return ($A, $B);
+}
+
 sub consume {
     my ($self, $rule, %parms) = @_;
     my $root = $parms{root} || '/';
@@ -135,11 +160,11 @@ sub consume {
             }
 
             my $tag = $id_class eq '' ? $name || '*' : '*';
-            
+
             if (defined $parms{prefix} and not $tag =~ /[*:|]/) {
                 $tag = join ':', $parms{prefix}, $tag;
             }
-            
+
             if (! $wrote_tag++) {
                 push @parts, $tag;
             }
@@ -179,7 +204,7 @@ sub consume {
                 my( $new_parts, $leftover ) = $self->consume( $rule, %parms );
                 shift @$new_parts; # remove '//'
                 my $xpath = join '', @$new_parts;
-                
+
                 push @parts, "[not(self::$xpath)]";
                 $rule = $leftover;
             }
@@ -191,6 +216,14 @@ sub consume {
         while ($rule =~ s/$reg->{pseudo}//) {
             if ( my @expr = $self->parse_pseudo($1, \$rule) ) {
                 push @parts, @expr;
+            } elsif ( $1 eq 'disabled') {
+                push @parts, '[@disabled]';
+            } elsif ( $1 eq 'checked') {
+                push @parts, '[@checked]';
+            } elsif ( $1 eq 'selected') {
+                push @parts, '[@selected]';
+            } elsif ( $1 eq 'text') {
+                push @parts, '*[@type="text"]';
             } elsif ( $1 eq 'first-child') {
                 # Translates to :nth-child(1)
                 push @parts, nth_child(1);
@@ -198,24 +231,30 @@ sub consume {
                 push @parts, nth_last_child(1);
             } elsif ( $1 eq 'only-child') {
                 push @parts, nth_child(1), nth_last_child(1);
-            } elsif ($1 =~ /^lang\(([\w\-]+)\)$/) {
+            } elsif ($1 =~ /^lang\($/) {
+                $rule =~ s/\s*([\w\-]+)\s*\)//
+                    or Carp::croak "Can't parse language part from $rule";
                 push @parts, "[\@xml:lang='$1' or starts-with(\@xml:lang, '$1-')]";
-            } elsif ($1 =~ /^nth-child\(odd\)$/) {
-                push @parts, nth_child(2, 1);
-            } elsif ($1 =~ /^nth-child\(even\)$/) {
-                push @parts, nth_child(2, 0);
-            } elsif ($1 =~ /^nth-child\((\d+)\)$/) {
-                push @parts, nth_child($1);
-            } elsif ($1 =~ /^nth-child\((\d+)n(?:\+(\d+))?\)$/) {
-                push @parts, nth_child($1, $2||0);
-            } elsif ($1 =~ /^nth-last-child\((\d+)\)$/) {
-                push @parts, nth_last_child($1);
-            } elsif ($1 =~ /^nth-last-child\((\d+)n(?:\+(\d+))?\)$/) {
-                push @parts, nth_last_child($1, $2||0);
-            } elsif ($1 =~ /^first-of-type$/) {
-                push @parts, "[1]";
-            } elsif ($1 =~ /^nth-of-type\((\d+)\)$/) {
-                push @parts, "[$1]";
+            } elsif ($1 =~ /^nth-child\(\s*$/) {
+                my( $A, $B ) = consume_An_plus_B(\$rule);
+                push @parts, nth_child($A, $B);
+            } elsif ($1 =~ /^nth-last-child\(\s*$/) {
+                my( $A, $B ) = consume_An_plus_B(\$rule);
+                push @parts, nth_last_child($A, $B);
+            } elsif ($1 =~ /^first-of-type\s*$/) {
+                my $type = $parts[-1];
+                push @parts, _generate_child('preceding', $type, 0, 1);
+
+            } elsif ($1 =~ /^nth-of-type\(\s*$/) {
+                my( $A, $B ) = consume_An_plus_B(\$rule);
+                my $type = $parts[-1];
+                push @parts, _generate_child('preceding', $type, $A, $B);
+
+            } elsif ($1 =~ /^last-of-type$/) {
+                push @parts, "[last()]";
+
+            # Err?! This one does not really exist in the CSS spec...
+            # Why did I add this?
             } elsif ($1 =~ /^contains\($/) {
                 if( $rule =~ s/^\s*"([^"]*)"\s*\)// ) {
                     push @parts, qq{[text()[contains(string(.),"$1")]]};
@@ -268,7 +307,7 @@ sub to_xpath {
     my $self = shift;
     my $rule = $self->{expression} or return;
     my %parms = @_;
-    
+
     my($result,$leftover) = $self->consume( $rule, %parms );
     $leftover
         and die "Invalid rule, couldn't parse '$leftover'";
@@ -276,9 +315,9 @@ sub to_xpath {
 
 }
 
-sub parse_pseudo { 
+sub parse_pseudo {
     # nop
-}    
+}
 
 1;
 __END__
@@ -317,7 +356,8 @@ CSS2 and partial CSS3 selectors to the equivalent XPath expression.
 
   $xpath = selector_to_xpath($selector, %options);
 
-Shortcut for C<< HTML::Selector->new(shift)->to_xpath(@_) >>. Exported upon request.
+Shortcut for C<< HTML::Selector::XPath->new(shift)->to_xpath(@_) >>.
+Exported upon request.
 
 =item new
 
@@ -345,10 +385,10 @@ prefix for the generated XPath expression.
 
 =item parse_pseudo
 
-This method is called during xpath construction when we encounter a pseudo 
-selector (something that begins with comma). It is passed the selector and 
-a reference to the string we are parsing. It should return one or more 
-xpath sub-expressions to add to the parts if the selector is handled, 
+This method is called during xpath construction when we encounter a pseudo
+selector (something that begins with comma). It is passed the selector and
+a reference to the string we are parsing. It should return one or more
+xpath sub-expressions to add to the parts if the selector is handled,
 otherwise return an empty list.
 
 =back
@@ -399,5 +439,6 @@ it under the same terms as Perl itself.
 
 L<http://www.w3.org/TR/REC-CSS2/selector.html>
 L<http://use.perl.org/~miyagawa/journal/31090>
+L<https://en.wikibooks.org/wiki/XPath/CSS_Equivalents>
 
 =cut
